package nxm.database.lib;


import  java.sql.*;

public class MySQLDB extends NxmDB {
  public static final int defaultPort = 3306;

  public MySQLDB() {
    super();
    if (myPort == 0) myPort = defaultPort;
  }

  public MySQLDB(String db, String user, String passwd) {
    super("localhost", defaultPort, db, user, passwd);
  }

  public MySQLDB(String host, String db, String user, String passwd) {
    super(host, defaultPort, db, user, passwd);
  }

  public MySQLDB(String host, int port, String db, String user, String passwd) {
    super(host, port, db, user, passwd);
  }

  public void connect() throws DatabaseException {
    String conStr;
    try {
      myDriver = (Driver)Class.forName("org.gjt.mm.mysql.Driver").newInstance();
      conStr = "jdbc:mysql://" + myHost + ":" + myPort + "/" + myName;
      myConnection = DriverManager.getConnection(conStr, myUserName, myUserPassword);

      if (verbose) {
        SQLWarning warn = myConnection.getWarnings();
        while (warn != null) {
          System.out.println("SQLState = " + warn.getSQLState());
	  System.out.println("Message  = " + warn.getMessage());
	  System.out.println("Vendor   = " + warn.getErrorCode());
	  System.out.println("");
	  warn = warn.getNextWarning();
        }
      }
    } catch (Exception e) {
      throw(new DatabaseException(e.getMessage()));
    }

   }

  public ResultSet getTables() {
    ResultSet rs;
    try {
      Statement stmt = myConnection.createStatement();
      rs = stmt.executeQuery("SHOW TABLES;");
    } catch (SQLException e) {
      if (verbose) { System.out.println("getTables(): " + e);}
      rs = null;
    }
    return(rs);
  }

  public ResultSet describeTable(String t) {
    ResultSet rs;
    try {
			rs = myConnection.getMetaData().getColumns(null, null, t, null);
    } catch (SQLException e) {
      if (verbose) { System.out.println("getTables(): " + e);}
      rs = null;
    }
    return (rs);
  }

}
