package nxm.nmpostlegacy.lib;


import java.net.Socket;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import nxm.sys.lib.Midas;
import nxm.sys.lib.DataFile;



/*
 * @Title:        SendFile</p>
 * @Description: This primitive POSTS XML multipart Files
 * @Company:    Zeta Associates</p>
 * @author        Wayne G. Dengel II
 * @version 1.0
 */

public class SendFile implements Runnable {
	private final static int 	BUFSIZE			= 8192;

	private Midas				M				= null;
	private boolean 			deBug			= false;						// GLOBAL Debug Flag
	private Socket 				clientSocket	= null;						// The Socket Reference
   	private Thread				mySelf			= null;						// Thread instance so we can start ourselves...
	private String 				host			= null;						// Host to connect to
	private int					port			= 0;						// Port to connect to...
	private String				filename		= null;						// The File we are about to send...
	private int 				fsize			= 0;

  	private DataOutputStream 	dos				= null;						// Stream OUTPUT Buffer
	private FileInputStream 	fis 			= null;

	private int					status			= 0;						// Status of Transmission...
	public final static int		COMPLETE		= 50;						// Transmission GOOD..
	public final static int		NOCONNECT		= 30;						// Failure to CONNECT/Create Streams
	public final static int		BADPARSE		= 29;						// Failure to PARSE the Table..
	public final static int		NOSEND			= 28;
	public final static int		ACKTIMEOUT		= 27;						// Failure to get ACK back from Recieving client

	private SendFile_cb			sf_callback		= null;

  	public SendFile(Midas M,String host, int port, String filename, String userpath,SendFile_cb sfCB){
		this.M 	  		= M;
		this.host 		= host;
		this.port 		= port;
		this.filename   = filename;
		sf_callback 	=  sfCB;
		mySelf = new Thread(this);
		mySelf.start();
  	}

	public void registerCallBack(SendFile_cb sfCB){
		sf_callback =  sfCB;
	}



  	public void run() {
		if(deBug)	info(" Thread Start "+filename);

		DataFile df = new DataFile(M,filename);
		try{
			df.open();
		}catch(Exception ex){
			info("DF.OPEN "+ex);
			if(sf_callback!=null){
				sf_callback.sfStatus(BADPARSE);
			}
			return;
		}

		String path = df.getURL();
		int i = path.indexOf(":/");
		if(i>-1){
			path = path.substring(i+1);
			if(deBug) info("Path = "+path);
		}else{
			if(sf_callback!=null){
				sf_callback.sfStatus(NOSEND);
			}
			return;
		}

		if(!openConnection()){
			if(deBug)	info(" Failure Openning Connection ");
			status = NOCONNECT;
			if(sf_callback!=null){
				sf_callback.sfStatus(NOCONNECT);
			}
			return;
		}

		if(!sendMultiPartStart()){
			if(deBug)	info(" Failure to sendMultiPartStart");
			status = NOSEND;
			closeConnection();
			if(sf_callback!=null){
				sf_callback.sfStatus(NOSEND);
			}
			return;
		}

		try{
			fis = new FileInputStream(path);
			fsize = fis.available();
            System.out.println("File size is "+fsize);

		}catch(Exception ex){
			info("Exception in file "+ex);
			if(sf_callback!=null){
				sf_callback.sfStatus(NOSEND);
			}
			return;
		}

		if(!sendXMLpart()){
			if(deBug)	info(" Failure sendXMLpart");
			status = NOSEND;
			closeConnection();
			if(sf_callback!=null){
				sf_callback.sfStatus(NOSEND);
			}
			return;
		}

		if(!sendFileData()){
			if(deBug)	info(" Failure sendFileData");
			status = NOSEND;
			closeConnection();
			if(sf_callback!=null){
				sf_callback.sfStatus(NOSEND);
			}
			return;
		}

		//try{
		//	Thread.sleep(5000);
		//}catch(Exception ex){}

		closeConnection();

		if(sf_callback!=null){
			sf_callback.sfStatus(COMPLETE);
		}

		status = COMPLETE;

	}

	private boolean sendFileData(){
		String  charbuf = null;
		byte [] dbuf 	= new byte[BUFSIZE];
		int 	ngot	= 0;

		if(deBug) info("in --- sendFileData\n");
		charbuf = new String("--blue1\r\n"+
							"Content-type: binary\r\n"+
							"Content-Description: Blue File\r\n"+
							"\r\n");

		if (!sendData(charbuf)) {
			info("out --- sendFileData --  1 Could not send to socket...");
			return false;
		}

		try{
            int i = 0;
			while(fis.available()>0){												/* Send DATA Here				*/
				ngot = fis.read(dbuf);
                i = i + ngot;
				if(ngot>0){
                    dos.write(dbuf,0,ngot);

					//String xxx = new String(dbuf);
					//if (!sendData(xxx)) {
					//	info("out --- sendFileData --  1 Could not send to socket...");
					//	return false;
					//}
				}
			}

            System.out.println("Sent "+i);

			fis.close();
		}catch(Exception ex){
			info("Excetpion sending sendFileData");
			return false;

		}
		if (!sendData(new String("\r\n--blue1--\r\n\r\n"))) {
			info("out --- sendFileData --  1 Could not send to socket...");
			return false;
		}
		if(deBug) info("out -- sendFileData\n");
		return true;
	}

	private boolean sendXMLpart(){
		String charbuf = null;
		String xmlstr  = new String();
		if(deBug) info("in --- sendXMLpart "+filename);
		xmlstr = xmlstr + "<XML>";

		if(fsize>0){
			xmlstr = xmlstr + "<STATUS=COMPLETE/>";
		}else{
			xmlstr = xmlstr + "<STATUS="+status+"/>";
		}
		xmlstr = xmlstr + "<USERIDENT="+filename+"/>";

		xmlstr = xmlstr + "<USERPATH="+"AUX"+"/>";
		xmlstr = xmlstr + "<FILESIZE="+fsize+"/></XML>";
		charbuf = "--blue1\r\n" + "Content-type: text/xml\r\n";

		if (!sendData(charbuf)) {
			info("out --- sendXMLpart --  1 Could not send to socket...");
			return false;
		}

		if (!sendData(new String("\r\n"))) {
			info("out --- sendXMLpart --  1.5 Could not send to socket...");
			return false;
		}

		if (!sendData(xmlstr)) {
			info("out --- sendXMLpart --  2 Could not send to socket...");
			return false;
		}
		if (!sendData(new String("\r\n--blue1--\r\n"))) {
			info("out --- sendXMLpart --  3 Could not send to socket...");
			return false;
		}
		if(deBug) info("out -- sendXMLpart\n");
		return true;
	}





	private boolean sendMultiPartStart(){
    	String str = new String();
		if(deBug) info("in --- sendMultiPartStart\n");
		str = str + "Content-type: Multipart/Related; boundary=blue1\r\n";
		str = str + "type=Application/plot\r\n\r\n";

		if (!sendData(str)) {
			info("out --- sendMultiPartStart --  Could not send to socket...");
			return false;
		}
		if(deBug) info("out -- sendMultiPartStart");
		return true;
	}
	private boolean sendData(String msgString){
		try{
			int msgSize = msgString.length();					// Get number of chars in STRING..
			if(deBug) info("SENDDATA "+msgSize);
			dos.write(msgString.getBytes());					// Write the Message to socket
			dos.flush();										// Make sure DATA is flushed..
			if(deBug) info("Tried to write "+msgSize+"  and wrote "+dos.size());
		}catch(Exception ex){
			info("Some exception writing socket "+ex);
			return false;
		}
		return true;
	}
	private boolean openConnection(){
		try{
			clientSocket = new Socket(host,port);
			dos	= new DataOutputStream(clientSocket.getOutputStream());
		}catch(Exception ex){
			info("openConnection: "+ex);
			return false;
		}
		return true;
	}
	private void closeConnection(){
		try{
			clientSocket.close();
		}catch(Exception ex){
			info("closeConnection: "+ex);
		}

	}
	public int getStatus(){
		return status;
	}
	private void info(String str){
		System.out.println("SendFile: "+str);
	}
}
