package nxm.nmpostlegacy.lib;

import java.net.Socket;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import nxm.sys.lib.Time;
import nxm.sys.lib.Table;

/**
   Title:       SendMsg <br>
   Description: This primitive POSTS XML NeXtMidas Messages to an ADDRESS/PORT.<br>
   <pre>
      The Input is in form of a TABLE...
           In table is:
      NAME    :   The Message Name itself
      TO      :   ID of NeXtMidas client to receive the message
      FROM    :   ID of the sender
      INFO    :   INFO Field...
      DATA    :   another Nested TABLE containing the data items.
                 This field CAN be omitted

      All Fields are MANITORY unless noted!!

     Please note the this only sends VALUES, it will not send objects..
         It parses the INPUT into XML..
         Maybe in future, I will serialize the OBJECTS so I can send them...
   </pre>
   Company:    Zeta Associates
   @author     Wayne G. Dengel II
   @version 1.0
   @version $Id: SendMsg.java,v 1.2 2013/01/07 21:09:41 ntn Exp $
 */
public class SendMsg implements Runnable {
  private boolean          deBug        = false; // GLOBAL Debug Flag
  private Socket           clientSocket = null;  // The Socket Reference
  private Thread           mySelf       = null;  // Thread instance so we can start ourselves...
  private String           host         = null;  // Host to connect to
  private int              port         = 0;     // Port to connect to...
  private Table            msgD         = null;  // The TABLE that hold message info
  private String           msgString    = null;  // The TABLE Parsed into XML stored in msgString
  private DataInputStream  dis          = null;  // Stream INPUT Buffer
  private DataOutputStream dos          = null;  // Stream OUTPUT Buffer
  private int              status       = 0;     // Status of Transmission...
  public static final int  COMPLETE     = 50;    // Transmission GOOD..
  public static final int  NOCONNECT    = 30;    // Failure to CONNECT/Create Streams
  public static final int  BADPARSE     = 29;    // Failure to PARSE the Table..
  public static final int  NOSEND       = 28;
  public static final int  ACKTIMEOUT   = 27;    // Failure to get ACK back from Receiving client
  private SendMsg_cb       sm_callback  = null;

  public SendMsg(String host, int port, Table msgD,SendMsg_cb smCB) {
    this.host   = host;
    this.port   = port;
    this.msgD   = msgD;
    sm_callback = smCB;
    mySelf = new Thread(this);
    mySelf.start();
  }

  public void run() {
    if (deBug) info(" Thread Start ");
    if (!openConnection()) {
      if (deBug) info(" Failure Openning Connection ");
      status = NOCONNECT;
      if (sm_callback!=null) {
        sm_callback.smStatus(NOCONNECT);
      }
      return;
    }
    if (!parseMsg()) {
      if (deBug) info(" Bad Parse");
      status = BADPARSE;
      closeConnection();
      if (sm_callback!=null) {
        sm_callback.smStatus(BADPARSE);
      }
      return;
    }

    if (!sendMsg()) {
      if (deBug) info(" Failure to SendMessage");
      status = NOSEND;
      closeConnection();
      return;
    }

    if (!waitAck()) {
      if (deBug) info(" Failure toACK");
      status = ACKTIMEOUT;
      closeConnection();
      if (sm_callback!=null){
        sm_callback.smStatus(ACKTIMEOUT);
      }
      return;
    }
    status = COMPLETE;
    if(sm_callback!=null){
      sm_callback.smStatus(COMPLETE);
    }
  }

  private boolean sendMsg() {
    try{
      int msgSize = msgString.length();  // Get number of chars in STRING..
      dos.write(msgString.getBytes());   // Write the Message to socket
      dos.flush();                       // Make sure DATA is flushed..
      if (deBug) info("Tried to write "+msgSize+"  and wrote "+dos.size());
      if (msgSize!=dos.size()) {
        if(deBug) info("Tried to write "+msgSize+"  and wrote "+dos.size());
        return false;
      }
    } catch(Exception ex) {
      info("Some exception writing socket "+ex);
      return false;
    }
    return true;
  }

  private boolean waitAck(){
    double t1 = Time.current();
    double t2 = 0.0;
    boolean waiting = true;
    String str1 = "HTTP/1.1 200 OK\r\nServer: NeXtMidas\r\nConnection: close\r\nContent-Type: text/xml\r\n\r\n";
    String str2 = "<XML><RESPONCE=\"YES\"/></XML>";
    String strfinal = str1 + str2;

    int waitNum = strfinal.length();
    waitNum--;
    if (deBug) info("waitAck:  Waiting for "+waitNum);

    try {
      while (waiting) {
        int avail = dis.available();
        if (avail>=waitNum) {
          return true;
        } else {
          if (deBug) info("Got "+avail);
        }
        t2 = Time.current();
        if ((t2-t1)>5.0) {
          info("Timeout Reached...");
          return false;
        }
        Thread.sleep(100);
      }
    } catch (Exception ex) {
      info("UnKnown Exception in waitAck :"+ex);
      return false;
    }
    return false;
  }

  private boolean parseMsg() {
    String nameField = null;
    String toField   = null;
    String fromField = null;
    String infoField = null;
    Table  dataT     = null;

    try {
      nameField = msgD.getS("NAME");
    } catch (Exception ex) {
      info("Unable to parse NAME field from table \n"+ex);
      return false;
    }

    try {
      toField = msgD.getS("TO");
    } catch (Exception ex) {
      info("Unable to parse TO field from table \n"+ex);
      return false;
    }

    try {
      fromField = msgD.getS("FROM");
    } catch(Exception ex) {
      info("Unable to parse FROM field from table \n"+ex);
      return false;
    }

    try {
      infoField = msgD.getS("INFO");
    } catch (Exception ex) {
      info("Unable to parse INFO field from table \n"+ex);
      return false;
    }
    try {
      dataT = msgD.getTable("DATA");
    } catch (Exception ex) {
      info("Unable to parse DATA sub-table from table \n"+ex);
      return false;
    }


    if ((nameField==null)||(nameField.equals(""))) {
      info("NAME in msg table is BLANK >"+nameField+"<");
      return false;
    }

    if ((toField==null)||(toField.equals(""))) {
      info("TO in msg table is BLANK >"+toField+"<");
      return false;
    }

    if ((fromField==null)||(fromField.equals(""))) {
      info("FROM in msg table is BLANK >"+fromField+"<");
      return false;
    }

    if ((infoField==null)||(infoField.equals(""))) {
      info("INFO in msg table is BLANK >"+infoField+"<");
      return false;
    }

    msgString = "POST / HTTP\r\nContent-type: text/plain\r\n\r\n<XML>";
    msgString = msgString +"<MSG NAME=\""+nameField+"\">";
    msgString = msgString +"<INFO=\""+infoField+"\"/>";
    msgString = msgString +"<TO=\""+toField+"\"/><FROM=\""+fromField+"\"/>";

    //msgString = "POST / HTTP\r\nContent-type: text/plain\r\n\r\n<XML>";
    //msgString = msgString +"<MSG NAME=\""+nameField+"\"";
    //msgString = msgString +" INFO=\""+infoField+"\"";
    //msgString = msgString +" TO=\""+toField+"\" FROM=\""+fromField+"\">";

    msgString = msgString +"<DATA> ";

    StringBuilder dataItems = new StringBuilder();
    if (dataT!=null) {
      dataT.enumeration();
      boolean first = true;
      while (dataT.more()) {
        String key = dataT.next();
        first = false;
        String value = dataT.getS(key);
        dataItems.append(key).append('=').append(value).append(',');
      }
    }
    msgString = msgString + dataItems + "</DATA></MSG></XML>";
    if (deBug) info("String is \n"+msgString);
    return true;
  }

  private boolean openConnection() {
    try {
      clientSocket = new Socket(host,port);
      dis = new DataInputStream(clientSocket.getInputStream());
      dos = new DataOutputStream(clientSocket.getOutputStream());
    } catch(Exception ex) {
      info("openConnection: "+ex);
      return false;    }
    return true;
  }

  private void closeConnection() {
    try {
      clientSocket.close();
    } catch(Exception ex) {
      info("closeConnection: "+ex);
    }
  }

  public int getStatus() {
    return status;
  }

  private void info(String str) {
    System.out.println("SendMsg: "+str);
  }
}
