package nxm.nmpostlegacy.prim;

import nxm.sys.lib.*;

import java.util.*;
import java.text.DecimalFormat;

/**
  conv_time
  Takes J1950 time, and converts from YYYY:MM:DD::HH:MM:SS.XXX to HH:MM:SS.XXX
  and places it into a RESULT
  Converts j1950 Time to Date/Time String

  @author Wayne G. Dengel II
  @version %I%, %G%
 */
public class conv_time extends Primitive {
  public int open() {
    double   inTime  = MA.getD("INTIME");
    String   timeRes = MA.getS("TIME","TIME");    // TIME
    String   dateRes = MA.getS("DATE","DATE");    // Date
    Calendar cal     = null;
    String   tTime   = null;
    String   tTime2  = null;
    String   tDate   = null;

    //M.info("CONV_TIME:  intime="+inTime);

    //String [] myzones = TimeZone.getAvailableIDs();

    //for (int i=0;i<myzones.length;i++){
    //  M.info("..."+myzones[i]);
    //}

    if (inTime == 0.0) {
      TimeZone myzone = new SimpleTimeZone(0,"UTC");

      cal = new GregorianCalendar(myzone);
      DecimalFormat decFormat = new DecimalFormat();

      decFormat.setMinimumIntegerDigits(2);

      //tTime2 = new String(cal.get(Calendar.HOUR_OF_DAY)+":"+
      //          cal.get(Calendar.MINUTE)+":"+
      //          cal.get(Calendar.SECOND));
      tTime2 =
          decFormat.format(cal.get(Calendar.HOUR_OF_DAY)) +":"+
              decFormat.format(cal.get(Calendar.MINUTE)) +":"+
              decFormat.format(cal.get(Calendar.SECOND));


      tDate  =
          cal.get(Calendar.YEAR)+":"+
              decFormat.format(cal.get(Calendar.MONTH)+1)+":"+
              decFormat.format(cal.get(Calendar.DAY_OF_MONTH));

    } else {
      tTime     = timeString(inTime);
      //M.info("Converted Time is "+tTime+" from "+inTime);
      int index   = tTime.indexOf("::");
      tTime2     = tTime.substring(index+2);
      tDate       = tTime.substring(0,index);
    }

    //M.info(""+timeRes+" = "+tTime2);
    //M.info(""+dateRes+" = "+tDate);

    MR.put(timeRes,tTime2);
    MR.put(dateRes,tDate);
    return (FINISH);
  }

  public static String timeString(Calendar calendar) {
    DecimalFormat decFormat = new DecimalFormat();
    decFormat.setMinimumIntegerDigits(2);

    String str = Integer.toString(calendar.get(Calendar.YEAR)) + ":"
        + decFormat.format(calendar.get(Calendar.MONTH) + 1) + ":"
        + decFormat.format(calendar.get(Calendar.DATE)) + "::"
        + decFormat.format(calendar.get(Calendar.HOUR_OF_DAY)) + ":"
        + decFormat.format(calendar.get(Calendar.MINUTE)) + ":"
        + decFormat.format(calendar.get(Calendar.SECOND)) + ".";

    decFormat.setMinimumIntegerDigits(3);
    str = str + decFormat.format(calendar.get(Calendar.MILLISECOND));

    return str;
  }

  public static String timeString(double time) {
    String str;
    if (time > 86400) {
      Calendar calendar = timeToCalendar(time);
      str = timeString(calendar);
    } else {
      str = Double.toString(precRound(time));
    }
    return(str);
  }

  public static Calendar timeToCalendar(double time) {
    int num_seconds = (int) time;
    double frac_sec = time - num_seconds;
    GregorianCalendar calendar = new GregorianCalendar(1950, 0, 1);
    calendar.setTimeZone(new SimpleTimeZone(0, "GMT"));
    calendar.add(Calendar.SECOND, num_seconds);
    calendar.add(Calendar.MILLISECOND, (int) (frac_sec * 1000));
    return(calendar);
  }

  public static double timeToMidas(String desiredTime) {
    StringTokenizer st = new StringTokenizer(desiredTime, ":");
    String strYear = null;
    String strMonth = null;
    String strDay = null;
    String strHour = null;
    String strMinute = null;
    String strSecond = null;
    try {
      strYear = st.nextToken();
      strMonth = st.nextToken();
      strDay = st.nextToken();
      strHour = st.nextToken();
      strMinute = st.nextToken();
      strSecond = st.nextToken();
    } catch (Exception e) {
      //System.out.println("TimeUtils:timeToMidas - invalid date/time string");
    }

    double theTime;

    if ( (strYear == null) || (strMonth == null) || (strDay == null) ||
        (strHour == null) || (strMinute == null) || (strSecond == null) ) {
      theTime = 0.0;
    } else {

      GregorianCalendar desiredCal = new GregorianCalendar(
          Integer.parseInt(strYear),
          Integer.parseInt(strMonth) - 1,
          Integer.parseInt(strDay),
          Integer.parseInt(strHour),
          Integer.parseInt(strMinute));
      desiredCal.setTimeZone(new SimpleTimeZone(0, "GMT"));
      Date desiredDate = desiredCal.getTime();
      long desiredSeconds = desiredDate.getTime() / 1000;

      GregorianCalendar epochCal = new GregorianCalendar(1950, 0, 1);
      epochCal.setTimeZone(new SimpleTimeZone(0, "GMT"));
      Date epochDate = epochCal.getTime();
      long epochSeconds = epochDate.getTime() / 1000;

      long secondsSinceEpoch = desiredSeconds - epochSeconds;

      theTime = secondsSinceEpoch + Double.parseDouble(strSecond);

      //System.out.println("TimeUtils:timeToMidas - desiredSeconds, epochSeconds, theTime = " + desiredSeconds + ", " + epochSeconds + ", " + theTime);
    }

    return(theTime);
  }

  public static double precRound(double val) {
    double round_val;
    long sign = 1;
    if (val < 0)
      sign = -1;

    round_val = ((long) (val*1000 + .5*sign)) / 1000.0;

    return round_val;
  }

}
