#!/bin/csh -f
#
# Configures for the X-Midas build. Note that this needs to link to the files in
# libxm that are built by NeXtMidas (see buildopt.txt and buildopt_libxm.txt).
#
# @author  whk
# @author  J. Czechowski IV
# @version $Id: nm.b,v 1.12 2011/07/22 14:51:08 ntn Exp $
#
#################################
if ($?JAVAHOME) then
  if ( $?XM_XFPPFLAGS ) then
    setenv XM_XFPPFLAGS "#XM${XM_VERSION}${XM_XFPPFLAGS}"
  else
    setenv XM_XFPPFLAGS "#XM${XM_VERSION}"
  endif
  xbh $1 bc

  # determine if on 32 or 64bit system and set jarch variable appropriately
  set arch = `uname -m`
  if ($arch == "x86_64") then
    set jarch = "amd64"
    set archbit = "64"
  else
    set jarch = "i386"
    set archbit = "32"
  endif

  setenv JNI_LIBS "-lawt -ljawt -ljvm"
  if ( -e ${JAVAHOME}/jre/lib/${jarch}/xawt/libmawt.so) then
    # Additional libmawt.so library in Java 5
    setenv JNI_LIBS "${JNI_LIBS} -lmawt"
  endif

  # NeXtMidas automatically adds the OS type to the end of the name for library
  # name when it is compiled (to allow a single installation to be compiled for
  # more than one OS). Although this is not possible when running via X-Midas
  # we still need to add the OS tag.
  if ( `uname` =~ "Linux" ) then
    setenv JNI_LIBS "${JNI_LIBS} -lxmjvm_lin${archbit}"
  else
    echo "ERROR: NM.B: Unknown OS '`uname`' can not determine lib name for libxmjvm"
    setenv JNI_LIBS "${JNI_LIBS} -lxmjvm"
  endif

  # Setup the rpath in the produced executable to use the
  # path to the Java JVM compiled against
  set old_xm_so_link_options = "$XM_SO_LINK_OPTIONS"
  set old_xm_so_link_libs = "$XM_SO_LINK_LIBS"
  set old_xm_exe_link_libs = "$XM_EXE_LINK_LIBS"
  set old_xm_exe_link_options = "$XM_EXE_LINK_OPTIONS"

  # Figure out if I am using the SUN jvm or the IBM jvm or one I do
  # not know about at all
  if ( -e ${JAVAHOME}/jre/lib/${jarch}/server/libjvm.so && -e ${JAVAHOME}/jre/lib/${jarch}/libjawt.so ) then
    #echo "Using SUN's Server JRE"
    setenv XM_EXE_LINK_OPTIONS "-Wl,-export-dynamic -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch} -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch}/native_threads -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch}/server -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch}/xawt -Wl,-rpath,${XMAREA_NXM}/libxm -L${JAVAHOME}/jre/lib/${jarch} -L${JAVAHOME}/jre/lib/${jarch}/native_threads -L${JAVAHOME}/jre/lib/${jarch}/server -L${JAVAHOME}/jre/lib/${jarch}/xawt -L${XMAREA_NXM}/libxm $XM_EXE_LINK_OPTIONS"
    setenv XM_EXE_LINK_LIBS "${JNI_LIBS} $XM_EXE_LINK_LIBS"
  else if ( -e ${JAVAHOME}/jre/lib/${jarch}/client/libjvm.so && -e ${JAVAHOME}/jre/lib/${jarch}/libjawt.so ) then
    #echo "Using SUN's Client JRE"
    setenv XM_EXE_LINK_OPTIONS "-Wl,-export-dynamic -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch} -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch}/native_threads -Wl,-rpath,${JAVAHOME}/jre/lib/${jarch}/client -Wl,-rpath,${XMAREA_NXM}/libxm -L${JAVAHOME}/jre/lib/${jarch} -L${JAVAHOME}/jre/lib/${jarch}/native_threads -L${JAVAHOME}/jre/lib/${jarch}/client -L${XMAREA_NXM}/libxm $XM_EXE_LINK_OPTIONS"
    setenv XM_EXE_LINK_LIBS "${JNI_LIBS} $XM_EXE_LINK_LIBS"
  else if ( -e ${JAVAHOME}/jre/bin/classic/libjvm.so && -e ${JAVAHOME}/jre/bin/libjawt.so ) then
    # IBM SDK 2
    echo   "**** Warning: You have selected the IBM SDK to use for Java"
    echo   "**** Note according to the DOCs of the IBMJAVA2-142"
    echo   "**** IBM does NOT currently support the JNI interface"
    echo   "**** This code is a placeholder for when it does"
    setenv XM_EXE_LINK_OPTIONS "-Wl,-export-dynamic -Wl,-rpath,${JAVAHOME}/jre/bin -Wl,-rpath,${JAVAHOME}/jre/bin/classic -Wl,-rpath,${XMAREA_NXM}/libxm -L${JAVAHOME}/jre/bin -L${JAVAHOME}/jre/bin/classic -L${XMAREA_NXM}/libxm $XM_EXE_LINK_OPTIONS"
    setenv XM_EXE_LINK_LIBS "-ljni ${JNI_LIBS} -lute -ljava $XM_EXE_LINK_LIBS"
  else
    echo "${0}: **** I do not know how to deal with your version of the JVM (${JAVAHOME})"
  endif

  # JC4: Hack to support X-Midas 4.2.1 (This might apply to XM 4.0.X too)
  if ($XM_VERSION >= 420) then
    setenv XM_EXE_LINK_LIBS "${XM_EXE_LINK_OPTIONS} ${XM_EXE_LINK_LIBS}"
  endif

  xbh $1 bl

  # Restore original link options
  setenv XM_SO_LINK_OPTIONS "$old_xm_so_link_options"
  setenv XM_SO_LINK_LIBS " $old_xm_so_link_libs"
  setenv XM_EXE_LINK_OPTIONS "$old_xm_exe_link_libs"
  setenv XM_EXE_LINK_LIBS " $old_xm_exe_link_options"
else
  echo "ERROR: Environment variable JAVAHOME not set, build aborted."
endif
