package nxm.sys.intr;

import nxm.sys.inc.*;
import nxm.sys.lib.*;
import nxm.sys.lib.Results;

/** Run a NeXtMidas command stored in a results string. This NeXtMidas intrinsic
    accepts the following flags to change the lexical scope (i.e. the results set)
    to use when running the command.
    <center><table border="1" summary="">
      <tr>
        <td align="center"><b>Flag</b></td>
        <td align="left"><b>Lexical Scope</b></td>
      </tr>
      <tr>
        <td align="center">&nbsp;<code>/GLOBAL</code>&nbsp;</td>
        <td align="left">Uses the global results set
                         (see {@link nxm.sys.lib.Results#getRoot()}).</td>
      </tr>
      <tr>
        <td align="center">&nbsp;<code>/PARENT</code>&nbsp;</td>
        <td align="left">Uses next results set up
                         (see {@link nxm.sys.lib.Results#getPrevious()}), if
                         available; otherwise uses the global results set.</td>
      </tr>
      <tr>
        <td align="center"><i>default</i></td>
        <td align="left">Uses the active results set.</td>
      </tr>
    </table></center>

    @author Jeff Schoen / J. Czechowski IV
    @version $Id: Run.java,v 1.12 2014/03/25 14:48:25 nmg Exp $
*/
public class Run extends Intrinsic {

  public int process () {
    Midas midas=M;

    if (MA.getState("/GLOBAL")) {
      midas = M.cloneOf();
      midas.results = (nxm.sys.lib.Results)MR.getRoot();
    }
    else if (MA.getState("/PARENT")) {
      midas = M.cloneOf();
      midas.results = (nxm.sys.lib.Results)MR.getPrevious();
      if (midas.results == null) {
        midas.results = (nxm.sys.lib.Results)MR.getRoot();
      }
    }

    String cmd = MA.getRawArgs().trim();
    int lcmd = cmd.length();

    if (lcmd>=2 && cmd.charAt(0)=='"' && cmd.charAt(lcmd-1)=='"') {
      cmd = cmd.substring(1, lcmd-1);
    }
    else if (Results.validNameLength(cmd,0)==lcmd) {
      String rcmd = MR.getS (cmd.toUpperCase());
      if (rcmd!=null) cmd = rcmd;
    }

    if (cmd.length()==0) M.warning ("RUN command given no arguments");
    else Shell.runCommand(midas, cmd);

    return FINISH;
  }

}
