package nxm.sys.lib;

import nxm.sys.inc.*;

import java.util.ArrayList;

/**
  A class to handle ESRI dbf (dBase III) files, which has Shapefile attribute format.

  @author Jeff Schoen
  @version $Id: DbfFile.java,v 1.19 2015/06/22 18:15:40 ntn Exp $
*/
public class DbfFile extends BaseFile implements ListFile {

  public static final int BUFSZ=32;

  private int fileCode,fileLength,fileOffset,recordSize,numFields,maxWidth;
  protected byte[] ghb = new byte[1024];

  private Field[] fields;

  public DbfFile () { }

  /**
       @deprecated since NeXtMidas 3.5.0 use {@link #DbfFile(MidasReference, Object)}
   */
  @Deprecated
  public DbfFile (Object ref, Object filename) {
    if (ref instanceof MidasReference) {
      init((MidasReference) ref, filename);
    } else {
      init(Convert.ref2Midas(ref), filename);
    }
  }

  /**
      @since NeXtMidas 3.5.0
   */
  public DbfFile (MidasReference ref, Object filename) {
    init(ref,filename);
  }

  /** short synopsis of file */
  public String toString () {
    String str = "DbfFile ";
    if (io!=null) str += io.getTypeString();
    str += " Resource";
    return str;
  }

  /** List the contents of the header */
  public String listHeader () {
    String list =
      "DbfFile   :  "+getURL()+"\n"
    + "Code        :  "+fileCode+"\n"
    + "Records     :  "+fileLength+"\n"
    + "RecSize     :  "+recordSize+"\n"
    + "DataOff     :  "+fileOffset+"\n"
    + "Size        :  "+getSize()+"\n"
    + "Comment     :  "+getComment()+"\n"
    + "Fields      :  "+numFields+"\n";
    for (int i=0; i<numFields; i++) {
      list += fields[i].toString()+"\n";
    }
    return list;
  }

  public int seek (double offset) {
    return super.seek((fileOffset+(offset*recordSize)+1) );
  }

  /** list a data element */
  public String listElements (double start, int elements, String format, int flags) {
    String line;
    byte[] db = new byte[recordSize];
    seek(start);
    int status = read(db,0,recordSize);
    if (status<=0) return null;
    line = "Record index: "+start+"\n";
    for (int i=0; i<numFields; i++) {
      Field f = fields[i];
      line +=  "    "+f.name+" = "+Convert.unpackS(db,f.start,f.length)+"\n";
    }
    return line;
  }
  public int listElementsPerLine (int lineWidth, String form, int flags) {
    return -numFields; // TODO why is this returning negative?
  }

  public String[] getElementArray (String elem) {
    return getElementArray(elem, 0, fileLength - 1);
  }

  public String[] getElementArray (String elem, int first, int last) {
    Field f = null;
    int i;
    byte[] db = new byte[recordSize];
    if (last < first) return (null);
    String[] el = new String[last - first + 1];
    double dfirst = (double)first;
    seek(dfirst);
    int recIndex = -1;
    for (i = 0; i < numFields; i++) {
      f = fields[i];
      if (f.name.equals(elem)) {
        recIndex = i;
	break;
      }
    }
    if (recIndex < 0) return (null);
    for (i = first; i <= last; i++) {
      read(db, 0, recordSize);
      String e = Convert.unpackS(db,f.start,f.length);
      el[i] = e;
    }

    return (el);
  }

  public String getTypeString () {
    String type="unknown";
    return type;
  }

  public boolean open () {
    boolean ok = super.open();
    if (!ok) return ok;
    byte[] hb = ghb;
    read (hb,0,BUFSZ);
    fileCode = (int)hb[0];
    fileLength = Convert.unpackL(hb, 4,EEEI);
    fileOffset = Convert.unpackI(hb, 8,EEEI);
    recordSize = Convert.unpackI(hb,10,EEEI);
    numFields = (fileOffset-BUFSZ-1)/BUFSZ;
    fields = getFields();
    return ok;
  }

  public Field[] getFields () {
    Field[] fields = new Field[numFields];
    maxWidth = 0;
    for (int i=0; i<numFields; i++) {
      fields[i] = new Field();
      maxWidth += fields[i].readFrom(this);
    }
    return fields;
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // ListFile Methods
  //////////////////////////////////////////////////////////////////////////////////////////////////
  public double getNumberOfRows () {
    return fileLength;
  }

  public boolean getProtected () {
    return true;
  }

  public Table getDataTable (double offset) {
    if (offset < 0) return null;
    Table  tbl = null;
    byte[] db  = new byte[recordSize];
    seek(offset);
    int status = read(db,0,recordSize);
    if (status > 0) {
      tbl = new Table();
      for (int i=0; i<numFields; i++) {
        Field f = fields[i];
        String str = Convert.unpackS(db,f.start,f.length);
        Object val = str;

        switch(Character.toUpperCase(f.type)) {
          case 'D': str = str.trim();
                    str = str.substring(0,4)+":"+str.substring(5,7)+":"+str.substring(7)+"::";
                    val = Double.valueOf(Convert.s2d(str));
                    break;
          case 'L': if (StringUtil.isTrue(str.trim().toUpperCase())) {
                      val = "TRUE";
                    } else if (StringUtil.isTrue(str.trim().toUpperCase())) {
                      val = "FALSE";
                    } else {
                      val = "?";
                    }
                    break;
          case 'N': val = Long.valueOf(Convert.o2x(str.trim()));   break;
          case 'F': val = Double.valueOf(Convert.s2d(str.trim())); break;
          case 'C': val = str; break;
          case 'M':
          case 'G':
          default:  val = str; break;
        }

        tbl.setKey(f.name, val);
      }
    }
    return tbl;
  }

  public int setData (double offset, Table table) {
    Shell.warning("DbfFile: setData(..) not implemented yet.");
    return -1;
  }

  public void setData (double element, String fieldName, Object obj) {
    Shell.warning("DbfFile: setData(..) not implemented yet.");
  }

  public void insertData (double offset, Object obj) {
    Shell.warning("DbfFile: insertData(..) not implemented yet.");
  }

  public final void removeData (double offset) {
    removeData(offset, 1L);
  }

  public void removeData (double offset, long count) {
    Shell.warning("DbfFile: removeData(..) not implemented yet.");
  }

  public int getRecordDefCount () {
    return numFields;
  }

  public void setRecordDefs (Table defs) {
    Shell.warning("DbfFile: getRecordDefs(..) not implemented yet.");
  }

  public Table getRecordDefs () {
    Table tbl = new Table();

    for(int i = 0; i < numFields; i++) {
      String fmt = "XA";
      tbl.setKey(fields[i].name, getRecordDef(i));

      // NTN: TODO: this seems fishy (fmt is not used, so why set it?), maybe
      //            the switch block below belongs in getRecordDef(int i)?
      switch(Character.toUpperCase(fields[i].type)) {
        case 'C': fmt = "XA"; break;       // ASCII Characters
        case 'D': fmt = "SD"; break;       // DATE in YYYYMMDD format
        case 'L': fmt = "1A"; break;       // LOGICAL
        case 'N': fmt = "SX"; break;       // (Integer) Number of up to 18 digits
        case 'F': fmt = "SD"; break;       // Float of up to 20 characters
        case 'M':                          // Pointer to MEMO block
        case 'G':                          // "General"
        default:  fmt = "XA"; break;
      }
    }

    return tbl;
  }

  public Table getRecordDef (int i) {
    Table tbl = null;

    if ((i >= 0) && (i < numFields)) {
      tbl = new Table();
      Field f = fields[i];
      tbl.setKey("NAME",   f.name  );
      tbl.setKey("FORMAT", f.format);
    }
    return tbl;
  }


  //////////////////////////////////////////////////////////////////////////////////////////////////
  // Nested Classes
  //////////////////////////////////////////////////////////////////////////////////////////////////
  public class Field {

    public String name,format;
    public char type;
    public int start,length,numdec;

    public int getLength () { return length; }

    public int readFrom (DbfFile sf) {
      byte[] hb=sf.ghb;
      sf.read (hb,0,BUFSZ);
      name = Convert.unpackS(hb,0,11);
      start = maxWidth;
      type = Convert.unpackC(hb,11);
      length = toUnsignedInt(Convert.unpackB(hb,16));
      numdec = toUnsignedInt(Convert.unpackB(hb,17));
      switch (type) {
        case 'C': case 'c': case 'D': case 'L': case 'M': case 'G':
          format = "XA";
          break;
        case 'N': case 'n': case 'F': case 'f':
          format = "SD";
          break;
        default:
      }
      return getLength();
    }

    /** Converts an unsigned byte into an int. */
    private int toUnsignedInt (byte val) {
      return ((int)val) & 0xFF;
    }

    public String toString() {
      return "Field Name="+name+" Type="+type+" Offset="+start+" Length="+length;
    }
  }

  public Record[] getRecords () {
    ArrayList<Record> list = new ArrayList<Record>(1024);
    for (; fileOffset<fileLength; ) {
      Record record = new Record();
      fileOffset += record.readFrom(this);
      list.add(record);
    }
    Record[] records = list.toArray(new Record[list.size()]);
    return records;
  }

  public class Record {

    public byte[] db = new byte[recordSize];

    public int getLength () {
      return recordSize;
    }

    public int readFrom (DbfFile sf) {
      sf.read (db,0,52);
      return getLength();
    }
  }
}
