package nxm.sys.lib;

import nxm.sys.inc.*;
import nxm.sys.libg.MWindow;
import nxm.sys.libg.Theme;

import java.awt.Container;

/**
  Base class for graphical primitives.  This is a container for methods that are
  specific to graphical primitives.

  They normally execute in a sub-thread of the current shell.

  @author Jeff Schoen
  @version $Id: GPrimitive.java,v 1.15 2015/03/16 17:22:06 ntn Exp $
*/
public class GPrimitive extends Command implements Chainable, Constants {

  static { Theme.initialize(); }  // added in 3.1.0 to make sure Graphical primitives pick up theme (Bug 2529)

  @InternalUseOnly("since 3.5.3, use getMWindow() or setMWindow")
  public MWindow MW; // this should become protected (or maybe even private) in 3.9+

  @Override
  public Object getNextLink() { return MW; }

  @Override
  public Object getPrevLink() { return null; }

  @Override
  public boolean isPipable() { return true; }

  @Override
  public void checkout() {
    super.checkout();

    if (MW != null) { MW.close(); }
    MW = null;
  }

  /** Gets the primary container used when displaying this graphical primitive. Subclasses may
      override this as necessary.
      @return The graphics panel or null if none is set.
      @since NeXtMidas 2.5.0
   */
  public Container getGraphicsContainer() {
    return (MW == null)? null: MW.panel;
  }

  /** @return MWindow for this graphical primitive
      @since NeXtMidas 3.5.3
   */
  public MWindow getMWindow () { return MW; }

  /** @param mw MWindow to use for this graphical primitive
      @since NeXtMidas 3.5.3
   */
  protected void setMWindow (MWindow mw) { MW = mw; }
}
