package nxm.sys.lib;

import nxm.sys.inc.*;

/** 
  Base class for native primitives (coded in C).  <br>
  <br>
  Native commands are primitives whose open(), process(), and close()
  methods are written entirely in C and compiled for the local machine
  and/or processing nodes on DSP cards.<br>
  <br>
  This base class is used to sequence the calls to those methods. 
  To build the C code, use nmbld.<br>
  <br>
  This is a container for methods that are specific to primitives.  
  These include data generators, data processors, and data sinks. <br>
  <br>
  They normally execute in a sub-thread of the current shell.

  @author Jeff Schoen
  @version $Id: NPrimitive.java,v 1.3 2004/07/26 13:46:39 jxc Exp $
*/
public class NPrimitive extends Command
		implements Constants {

  /** {@inheritDoc} */
  public int checkin (Midas M) { 
    return super.checkin(M);
  }

  /** {@inheritDoc} */
  public void setArgs (Args args) {
    super.setArgs(args);
  }                                                                             

  /** {@inheritDoc} */
  public int open() {
    return (NOOP);
  }
 
  /** {@inheritDoc} */
  public int process() {
    return (FINISH);
  }
 
  /** {@inheritDoc} */
  public int processMessage (Message msg) {
    return (NOOP);
  }
 
  /** {@inheritDoc} */
  public int close() {
    return (NOOP); 
  }
}
