package nxm.sys.libg;

import java.awt.Color;

/**
  A class to hold and plot a grid layer.

  @author Jeff Schoen
*/
public class LayerGrid extends Layer {

  private int dim;
  private float[] fbuf;

  /** Constructor */
  public LayerGrid () {
    super();
    setColorMap(MColor.GRID,-1);
    name = "GRID";
  }

  /** update View. */
  @SuppressWarnings("deprecation") // for usage of MPlot.viewLast until it becomes package-private in 3.7+
  public void update () {
    int n,nn,ic; double r,rmax;
    if (MP.view==MP.viewLast) return;
    ndata = 0;
    if (MP.view==MPlot.V_POLAR) {
      rmax = Math.max(Math.abs(MP.ory1),Math.abs(MP.ory2));
      fbuf = new float[1028*2];
      for (n=30; n<=90; n+=30) {
        ic=MColor.C_GRID; r = rmax*n/90;
        set2D (73,ic); set2D (-180.0,r); set2D (+180.0,r);
        for (nn=-180; nn<=180; nn+=5) set2D( (double)nn, r);
      }
      for (n=-180; n<=180; n+=30) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (19,ic); set2D (n,0); set2D (n,90);
        for (nn=0; nn<=90; nn+=5) set2D( (double)n, rmax*nn/90);
      }
      dim=2;
    }
    else if (MP.view==MPlot.V_LATLON || MP.view==MPlot.V_MERCATOR || MP.view==MPlot.V_GEODETIC) {
      fbuf = new float[4184*2];
      for (n=-75; n<=75; n+=15) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (181,ic); set2D (-180,n); set2D (180,n);
        for (nn=-180; nn<=180; nn+=2) set2D(nn,n);
      }
      for (n=-180; n<=165; n+=15) {
        if (n==0) ic=MColor.C_AXIS; else ic=MColor.C_GRID;
        set2D (87,ic); set2D (n,-85); set2D (n,85);
        set2D (n,-85); for (nn=-84; nn<=84; nn+=2) set2D(n,nn); set2D (n,85);
      }
      dim=2;
    }
    else if (MP.view==MPlot.V_XYZ) {
      int np=5,np2=np*2;
      fbuf = new float[3*(3+np2+1)*3];
      double xc=0.0, xd=1.0/np;
      double yc=0.0, yd=1.0/np;
      double zc=0.0, zd=1.0/np;
      ic=MColor.C_AXIS;
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc+np*xd,yc,zc);
      for (n=-np; n<=np; n++) set3D(xc+n*xd,yc,zc);
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc,yc+np*yd,zc);
      for (n=-np; n<=np; n++) set3D(xc,yc+n*yd,zc);
      set3D (np2+1,ic,0); set3D(xc,yc,zc); set3D(xc,yc,zc+np*zd);
      for (n=-np; n<=np; n++) set3D(xc,yc,zc+n*zd);
      ic=MColor.C_GRID;
      dim=3;
    }
  }

  private void set2D (int x, int y) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y;
  }
  private void set2D (double x, double y) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y;
  }
  private void set3D (int x, int y, int z) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; fbuf[ndata++]=(float)z;
  }
  private void set3D (double x, double y, double z) {
    fbuf[ndata++]=(float)x; fbuf[ndata++]=(float)y; fbuf[ndata++]=(float)z;
  }

  /** find layer bounds */
  @Override
  public void findRange() {
    super.findRange(); // start with default values
    if (MP.view==MPlot.V_LATLON || MP.view==MPlot.V_MERCATOR || MP.view==MPlot.V_GEODETIC) {
      x1=-180; x2=+180; y1=-90; y2=+90;
      // Set Minimum and Maximum actual auto-scale value (since 3.7.0 for NXM-2984)
      a1 = y1;
      a2 = y2;
    }
  }

  public void draw (int flag) {
    super.draw(flag);
    if (ndata<=0) return;

    int k;
    Line lin = line.copy();
    int dim   = this.dim;                // since 3.5.2 make local copy of dimension field for thread safety for BUG 2337
    int ndata = this.ndata;              // since 3.5.2 make local copy of ndata field for thread safety for BUG 2337
    float[] fbufRef = this.fbuf;         // since 3.5.1 b/c update() can change the fbuf underneath this draw() for BUG 2337
    final int fbufLen = fbufRef.length;  // since 3.5.1
    final Color[] colors = this.colors;  // since 3.5.2 to prevent ArrayIndexOutOfBoundsException for BUG 2337
    final int colorsLen = colors.length; // since 3.5.2
    for (int j=0; j<ndata && (j+2)<fbufLen; j=k) { // since 3.5.1 check that j is in array bounds
      int npts = (int)fbufRef[j++];
      int ncol = (int)fbufRef[j++];      // color number
      if (dim>2) j += (dim-2);
      k = j + (2+npts)*dim;
      // Check if next j will be invalid due to changing values (zoom in, zoom out) since 3.5.2
      if (k < 0) {
        // If this warning never occurs, due to thread safety practices implemented in 3.5.2,
        // the warning can be removed in 3.9.x
        warning("LayerGrid.draw invalid next j of " + k + ": current j=" + j + " npts=" + npts
            + " dim=" + dim + " fbufLen=" + fbufLen + " ndata=" + data );
        break;
      }
      if (MP.view==MPlot.V_GEODETIC || MP.view==MPlot.V_XYZ || MP.view==MPlot.V_POLAR) {
        j += 2*dim;
      } else {
        // since 3.5.2 check that j<fbufLen to prevent potential ArrayIndexOutOfBoundsException
        if ((j<fbufLen) && (fbufRef[j++]>=MP.rx2)) continue;
        if ((j<fbufLen) && (fbufRef[j++]>=MP.ry2)) continue;
        if ((j<fbufLen) && (fbufRef[j++]<=MP.rx1)) continue;
        if ((j<fbufLen) && (fbufRef[j++]<=MP.ry1)) continue;
      }
      if (j>=fbufLen) break;      // since 3.5.1 prevent potential ArrayIndexOutOfBoundsException for BUG 2337
      if (ncol < colorsLen && ncol > 0) { // since 3.5.2 to prevent ArrayIndexOutOfBoundsException for BUG 2337
        lin.setColor(colors[ncol]);
      } else {  // warn if ncol is negative and/or out of range  since 3.5.2 for BUG 2337
        warning("LayerGrid.draw color number " + ncol + " is Out Of Range, using previous color "+ lin.getColor());
      }
      if (MP.view==MPlot.V_XYZ) MP.pix.setFlag(Pix.NORM);
      MP.drawLine (fbufRef,j,dim,npts,lin);
      if (MP.view==MPlot.V_XYZ) MP.pix.clrFlag(Pix.NORM);
    }
  }

}
