package nxm.sys.libg;

import nxm.sys.lib.*;

import java.awt.image.*;

/** A class to hold and plot images from the NeXtMidas Map Server (NMS). Since NeXtMidas 2.5.0 this
    extends {@link LayerMap} for most of its capabilities.

    @author  J. Czechowski IV
    @version $Id: LayerNMS.java,v 1.21 2013/05/21 17:35:58 ntn Exp $
    @since   NeXtMidas 2.1.0
 */
public class LayerNMS extends LayerMap {
  private   static final String  PROTOCOL     = "NMS_1";        // Protocol id/version.
  private   static final boolean CHECK_LENGTH = true;           // Check image length?
  private   static final int     MAX_TRYS     = 3;              // Max retries

  protected String  options     = "DEF";   // Options (as given by user).
  protected String  optionsStr  = "DEF";   // Options (as sent to server).
  protected String  background  = "NONE";  // Background to use.

  /** Default constructor */
  public LayerNMS () {
    super("NMS");
    setDownloadThreads(8);
    warn("*********************************************************************");
    warn("** LayerNMS: This is a beta version of LayerNMS used to access the **");
    warn("**     NeXtMidas Map Server. This layer will change significantly  **");
    warn("**     before the next release.                                    **");
    warn("*********************************************************************");
  }

  protected boolean getDefUseTile ()     { return true; }
  protected int     getDefTileLevels ()  { return 32;   }
  protected int     getDefTileWidth ()   { return 256;  }
  protected int     getDefTileHeight ()  { return 256;  }

  public void setUseTiles (boolean tile) {
    if (!tile) warn("This service requires the use of tiles. UseTiles can not be changed to false.");
    super.setUseTiles(true);
  }

  public boolean isConnected() {
    return (getURL() != null);
  }

  /** Sets the name of the background to use. */
  public void setBackground(String bg) {
    if (!bg.equalsIgnoreCase(background)) {
      background = bg;
      refresh();
    }
  }

  /** Gets the name of the background to use. */
  public String getBackground() {
    return background;
  }

  /** Sets the options mask indicating what to display. */
  public void setOptions (String opts) {
    if (!opts.equalsIgnoreCase(options)) {
      String[] optItems = opts.toUpperCase().split("[|]");
      for (int i = 0; i < optItems.length; i++) {
              if (optItems[i].startsWith("-")) { optItems[i] = "NO"+optItems[i].substring(1);  }
         else if (optItems[i].startsWith("+")) { optItems[i] = "ADD"+optItems[i].substring(1); }
      }

      options    = opts;
      optionsStr = StringUtil.join(optItems, "+");
      refresh();
    }
  }

  /** Gets the options mask indicating what to display. */
  public String getOptions () {
    return options;
  }

  protected int drawArea(double minLon, double maxLon, double minLat, double maxLat,
                         int width, int height, boolean hiRes, String pos) {
    // Get and plot each of the tiles
    String url = mapServer+"?ver="+PROTOCOL+"&background="+background+"&options="+optionsStr+"&v=3&t="+pos;
    BufferedImage img = requestImage(url, MAX_TRYS, CHECK_LENGTH, width, height);
    if (img != null) drawImage(img, minLon, maxLon, minLat, maxLat);
    return 0;
  }

  public Table getConfig() {
    Table  tbl = super.getConfig();

    tbl.put("BACKGROUND", getBackground());
    tbl.put("OPTIONS",    getOptions());

    return tbl;
  }
}
