package nxm.sys.libg;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.awt.*;

/**
  A class to hold and plot a X vs Y data layer.

  @author Jon Grasmeder, Jay Hallam
  @version $Id: LayerXY.java,v 1.25 2014/12/11 15:55:12 ntn Exp $
*/
public class LayerXY extends Layer {

  /** The buffer to be drawn: (x1,y1,x2,y2,...xn,yn)*/
  public double[] dbuf;
  private int colorIndex;
  private int xOffset;
  private int yOffset;
  private char xFormat;
  private char yFormat;
  private String tag;
  private Color[] colors;

  private double xMin, yMin;
  private double xlMin, ylMin;
  private boolean xnotDouble = true;
  private boolean ynotDouble = true;

  /* Constructors */
  public LayerXY () {
    super();
    line.setColor(MColor.getColorByIndex(1));
    super.name = "XY";
  }

  /** Set x-record name: TBD */
  public void setXRecName (String value) {
    //System.err.println("LayerXY.setXRecName not implemented");
  }

  /** Set x-record name: shorthand keyable "x=" syntax */
  public void setX (String value) {
    setXRecName(value);
  }

  /** Set y-record name: TBD */
  public void setYRecName (String value) {
    //System.err.println("LayerXY.setYRecName not implemented");
  }

  /** Set y-record name: shorthand keyable "x=" syntax */
  public void setY (String value) {
    setYRecName(value);
  }

  /** setup the file input - if necessary */
  public boolean setFile (PlotFile pf, Table tab, Args MA) {
    setFile(pf);
    boolean returnStatus = false;
    int xferLength = (int) pf.getSize(); //MA.getL("TL",1);

    if (pf instanceof DataFile) { // JC4 20MAR06: This is the OLD way of doing it.
      DataFile hcb = (DataFile)pf;
      if (hcb.typeClass==3) {
        Object xRec = tab.getKey("X");
        Object yRec = tab.getKey("Y");
        if (xRec != null && yRec != null) {
          String xRecName = ""+xRec;
          String yRecName = ""+yRec;
          int xRecIndex = hcb.findRec(xRecName);
          int yRecIndex = hcb.findRec(yRecName);
          if (xRecIndex >= 0 && yRecIndex >= 0) {
            setColumns (xRecIndex, yRecIndex, hcb);
            setXAxis( 0, 1 ); //Initialize the x-axis start and delta
            setYAxis( 0, 1 ); //Initialize the y-axis start and delta
            Data dat = hcb.getDataBuffer( xferLength );
            setSize (dat.size);
            setData (dat);
          } else {
            warning("Unable to find X record '"+xRecName+"'" + " and/or Y record '"+yRecName+"' in header");
          }
        } else {
          warning("Need an X=TAG and Y=TAG for XY plot");
        }
      } else {
        warning("LayerXY plot for type 3000 file only");
      }
    } else if (pf instanceof ListFile) {
      ListFile lf = (ListFile)pf;

      String xRecName = tab.getS("X", "X");
      String yRecName = tab.getS("Y", "Y");

      tag = pf.getTag();
      super.name = tag;
      xOffset = 0;
      yOffset = 0;
      xFormat = ' ';
      yFormat = ' ';

      for (int i = 0; i < lf.getRecordDefCount(); i++) {
        Table  recDef = lf.getRecordDef(i);
        String fmt    = recDef.getS("FORMAT", "XA");

        if (xRecName.equals(recDef.getS("NAME"))) {
          xFormat = fmt.charAt(1);
          xOffset = recDef.getL("OFFSET", xOffset); // Type 3000 has OFFSET, others use xOffset
        }

        if (yRecName.equals(recDef.getS("NAME"))) {
          yFormat = fmt.charAt(1);
          yOffset = recDef.getL("OFFSET", yOffset); // Type 3000 has OFFSET, others use yOffset
        }

        if (xFormat == ' ') { xOffset = xOffset + Data.getBPA(fmt); }
        if (yFormat == ' ') { yOffset = yOffset + Data.getBPA(fmt); }
      }

      if (xFormat == ' ') {
        warning("LayerXY Can not find subrecord '"+xRecName+"' in "+this);
      }
      if (yFormat == ' ') {
        warning("LayerXY Can not find subrecord '"+yRecName+"' in "+this);
      }

      setXAxis( 0, 1 ); //Initialize the x-axis start and delta
      setYAxis( 0, 1 ); //Initialize the y-axis start and delta
      Data dat = pf.getDataBuffer(xferLength);
      setSize(dat.size);
      setData(dat);
    }

    return returnStatus;
  }

  public void process () {
    ndata = 0;
    if (data!=null) {
      ndata = data.size;
      dbuf = data2daXY (data,ndata,dbuf,MP.mode);
      setXAxis( xMin, 1 ); //jph -- hard-wired
      setYAxis( yMin, 1 ); //jph -- hard-wired
    }
  }

  private double convertData (Data data, int offset, char format) {
    double result = 0;
    switch (format) {
      case 'F':  result = data.unpackF(offset); break;
      case 'D':  result = data.unpackD(offset); break;
      case 'X':  result = data.unpackX(offset); break;
      case 'L':  result = data.unpackL(offset); break;
      case 'I':  result = data.unpackI(offset); break;
      case 'B':  result = data.unpackB(offset); break;
      default:   result = 0.0;                  break;
    }
    return result;
  }

  /** Convert the data into an (x1,y1,x2,y2,...xn,yn) array for plotting */
  private double[] data2daXY (Data data, int ndata, double[] dbuf, int mode) {
    double dbuftemp[] = new double[ndata];
    double scale   = setScale(data, mode);
    xMin = yMin = Double.MAX_VALUE;
    setxnotDouble(true);
    setynotDouble(true);

    // dbuf holds alternating x and y points
    if (dbuf==null||dbuf.length<ndata*2) dbuf = new double[ndata*2];

    if (xFormat == 'D') {
      for (int i=0; i < ndata; i++) dbuftemp[i] = data.unpackD(i*data.bpe + xOffset);
      for (int i=0; i < ndata; i++) if (dbuftemp[i]<xMin) xMin = dbuftemp[i];
      for (int i=0; i < ndata; i++) dbuf[i*2] = dbuftemp[i];
      setxnotDouble(false);
    }

    if (yFormat == 'D') {
      for (int i = 0; i < ndata; i++) dbuftemp[i] = data.unpackD(i*data.bpe + yOffset);
      for (int i = 0; i < ndata; i++) if (dbuftemp[i]<yMin) yMin = dbuftemp[i];
      for (int i = 0; i < ndata; i++) dbuf[i*2+1] = dbuftemp[i];
      setynotDouble(false);
    }

    if (getxnotDouble()) {
      for (int i = 0; i < ndata; i++)
        dbuf[i*2] = convertData(data, (i*data.bpe + xOffset), xFormat);
    }
    if (getynotDouble()) {
      for (int i = 0; i < ndata; i++)
        dbuf[i*2+1] = convertData(data, (i*data.bpe + yOffset), yFormat);
    }

    /** @todo this only scales the Y axis.  scale X axis as well? */
    switch (mode) {
      case MPlot.M_10LOG:
      case MPlot.M_20LOG:
        for (int i=0; i<ndata; i++) {
          int iod = i*2+1;
          int iev = i*2;
          dbuf[iod] = scale * Math.log(Math.max(10.e-11,Math.abs(dbuf[iod])));
        }
        break;
      default:
        break;
    }
    return dbuf;
  }

  /** find layer bounds - the x-min and y-min */
  public void findRange() {
    if (frame==0) frame=ndata;
    x1 = 0;
    x2 = 1;
    y1 = 0;
    y2 = 1;
    if (dbuf==null) return;
    x1 = x2 = dbuf[0];
    y1 = y2 = dbuf[1];
    for (int i=1; i<ndata; i++) {
      if (dbuf[i*2]  >x2) x2=dbuf[i*2];
      if (dbuf[i*2]  <x1) x1=dbuf[i*2];
      if (dbuf[i*2+1]>y2) y2=dbuf[i*2+1];
      if (dbuf[i*2+1]<y1) y1=dbuf[i*2+1];
    }
    if (MP.mode==MPlot.M_RVI) {  x1=y1; x2=y2; }

    autoScale4FindRange(y1, y2); // since 3.3.3 use helper method in super class

    setxlMin( xMin );
    setylMin( yMin );
  }

  public void draw (int flag) {
    super.draw(flag);
    if (ndata<=0) return;

    int i1=0;
    int i2=ndata;
    int i21;

    if (size>0) i2=frame;

    i21 = i2-i1;
    if (i21<=0) return;

    Line lin = line.copy();
    MP.drawLine (dbuf,0,2,data.size,lin);

  }

  public void setSize (int value) {
    super.setSize(value);
    if (size>1) {
      cmap = MColor.MULTI;
      colors = MColor.getColorMap(cmap,size);
      line.setColor(colors[1]); // jph -- TBD
    }
  }

  /** Sets the class variables: tag, name, xOffset, yOffset, xFormat, yFormat */
  public void setColumns (int x, int y, DataFile hcb) {
    tag     = hcb.getTag(); //jph -- 24mar04  + " ("+hcb.getRecName(x)+" v. "+hcb.getRecName(y)+")";
    super.name = tag;
    xOffset = hcb.getRecOffset(x);
    xFormat = (char) hcb.getRecFormatType(x);
    yOffset = hcb.getRecOffset(y);
    yFormat = (char) hcb.getRecFormatType(y);
  }

  public void setColorIndex (int value) {
    colorIndex = value;
  }

  public String getTag() {
    return tag;
  }

  public double getxMin() {
    return xMin;
  }

  public double getyMin() {
    return yMin;
  }

  public double getxlMin() {
    return xlMin;
  }

  public double getylMin() {
    return ylMin;
  }

  public void setxlMin (double x) {
    xlMin = x;
  }

  public void setylMin (double x) {
    ylMin = x;
  }

  public boolean getxnotDouble() {
    return xnotDouble;
  }

  public boolean getynotDouble() {
    return ynotDouble;
  }

  public void setxnotDouble (boolean x) {
    xnotDouble = x;
  }

  public void setynotDouble (boolean x) {
    ynotDouble = x;
  }

}
