package nxm.sys.libg;

import java.awt.*;
import java.awt.event.*;

import javax.swing.BorderFactory;
import javax.swing.JPasswordField;
import javax.swing.border.Border;

/** A class to handle a password text field. Extends JPasswordField to allow right mouse exit.
    @author Jeff Schoen / J Bauer
    @version $Id: MJPasswordField.java,v 1.2 2012/02/10 19:55:27 ntn Exp $
    @since neXtMidas 2.9.0
*/
public class MJPasswordField extends JPasswordField {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012021100311L; // for Serializable interface

  /** Constructor
      @param text The initial text to display
      @param columns Initial width
   */
  public MJPasswordField (String text, int columns) {
    super(text,columns);
    /* Specify the events of interest */
    enableEvents(AWTEvent.MOUSE_EVENT_MASK);
    enableEvents(AWTEvent.KEY_EVENT_MASK);
    Border outer = BorderFactory.createEtchedBorder(Theme.global().cwbh,Theme.global().cwbs );
    Border inner = BorderFactory.createEtchedBorder(Theme.global().cwbh,Theme.global().cwbs );
    super.setBorder(BorderFactory.createCompoundBorder(outer, inner));
    super.setEditable(true);
  }

  /** Overrides <CODE>java.awt.ProcessEvent(AWTEvent e)</CODE> in order to allow a right mouse exit.
      @param e The event.
   */
  public void processEvent (AWTEvent e) {
    if (e.getID()==MouseEvent.MOUSE_RELEASED &&
            (((InputEvent)e).getModifiers()&MouseEvent.BUTTON3_MASK)!=0) {
        processActionEvent(new ActionEvent(this,0,"EXIT"));
    }
    super.processEvent(e);
  }

  protected void processActionEvent(ActionEvent e) {
    ActionListener listener = super.getActionListeners()[0];
    if (listener != null) {
      listener.actionPerformed(e);
    }
  }

}
