package nxm.sys.libm;

import nxm.sys.inc.*;

/**
   Performs the Magnitude operation on arrays. <br>
   See {@link Operator} for a description of the available I/O Modes.
   @author  Auto-Generated
   @see     Operator
*/
@SuppressWarnings("cast")
public final class Magnitude extends Operator implements Constants {

  /** Create a new instance. */
  public Magnitude() {
    super();
  }

  /** Create a new instance.
      @param flags The processing flags to use.
   */
  public Magnitude (int flags) {
    super(flags);
  }

  /** Create a new instance.
      @param scale The scale factor to use.
   */
  public Magnitude (double scale) {
    super(scale);
  }

  /** Create a new instance.
      @param flags The processing flags to use.
      @param scale The scale factor to use.
   */
  public Magnitude (int flags, double scale) {
    super(flags,scale);
  }

  public final int validate (char type, int n, int mode) {
    switch (mode) {
      case SS: return n;
      case CS: return n;
      case VS: return n;
      default: return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(double[][],int,int)
   */
  public static final void SS (double[] a, double[] b, int n) {
      for (int i=0; i<n; i++) {
        b[i]=(double)Math.abs(a[i]);
      }
  }

  /** Apply this operator to mode {@link Operator#CS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(double[][],int,int)
   */
  public static final void CS (double[] a, double[] b, int n) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        b[i]=(double)Math.sqrt((double)(a[j]*a[j]+a[j+1]*a[j+1]));
      }
  }

  /** Apply this operator to mode {@link Operator#VS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(double[][],int,int)
   */
  public static final void VS (double[] a, double[] b, int n) {
      for (int i=0,k=0; i<n; i++,k+=3) {
        b[i]=(double)Math.sqrt((double)(a[k]*a[k]+a[k+1]*a[k+1]+a[k+2]*a[k+2]));
      }
  }

  public final int process (float[][] v, int n, int mode) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (float[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(float[][],int,int)
   */
  public static final void SS (float[] a, float[] b, int n) {
      for (int i=0; i<n; i++) {
        b[i]=(float)Math.abs(a[i]);
      }
  }

  /** Apply this operator to mode {@link Operator#CS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(float[][],int,int)
   */
  public static final void CS (float[] a, float[] b, int n) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        b[i]=(float)Math.sqrt((double)(a[j]*a[j]+a[j+1]*a[j+1]));
      }
  }

  /** Apply this operator to mode {@link Operator#VS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(float[][],int,int)
   */
  public static final void VS (float[] a, float[] b, int n) {
      for (int i=0,k=0; i<n; i++,k+=3) {
        b[i]=(float)Math.sqrt((double)(a[k]*a[k]+a[k+1]*a[k+1]+a[k+2]*a[k+2]));
      }
  }

  public final int process (int[][] v, int n, int mode) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (int[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SS: SS(v[0],v[1],n); return n;
      case CS: CS(v[0],v[1],n); return n;
      case VS: VS(v[0],v[1],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(int[][],int,int)
   */
  public static final void SS (int[] a, int[] b, int n) {
      for (int i=0; i<n; i++) {
        b[i]=(int)Math.abs(a[i]);
      }
  }

  /** Apply this operator to mode {@link Operator#CS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(int[][],int,int)
   */
  public static final void CS (int[] a, int[] b, int n) {
      for (int i=0,j=0; i<n; i++,j+=2) {
        b[i]=(int)Math.sqrt((double)(a[j]*a[j]+a[j+1]*a[j+1]));
      }
  }

  /** Apply this operator to mode {@link Operator#VS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param n     The number of elements to operate on.
      @see #process(int[][],int,int)
   */
  public static final void VS (int[] a, int[] b, int n) {
      for (int i=0,k=0; i<n; i++,k+=3) {
        b[i]=(int)Math.sqrt((double)(a[k]*a[k]+a[k+1]*a[k+1]+a[k+2]*a[k+2]));
      }
  }

  public int getInputs() { return 1; }
  public int getOutputs() { return 1; }
}
