package nxm.sys.libm;

import nxm.sys.inc.*;

/**
   Performs the ModuloEuclid operation on arrays. <br>
   Euclidean Modulo (the remainder of a/b) function (always return positive). <br>
   See {@link Operator} for a description of the available I/O Modes.
   @author  Auto-Generated
   @see     Operator
*/
@SuppressWarnings("cast")
public final class ModuloEuclid extends Operator implements Constants {

  /** Create a new instance. */
  public ModuloEuclid() {
    super();
  }

  /** Create a new instance.
      @param flags The processing flags to use.
   */
  public ModuloEuclid (int flags) {
    super(flags);
  }

  /** Create a new instance.
      @param scale The scale factor to use.
   */
  public ModuloEuclid (double scale) {
    super(scale);
  }

  /** Create a new instance.
      @param flags The processing flags to use.
      @param scale The scale factor to use.
   */
  public ModuloEuclid (int flags, double scale) {
    super(flags,scale);
  }

  public final int validate (char type, int n, int mode) {
    switch (mode) {
      case SSS: return n;
      default: return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (double[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(double[][],int,int)
   */
  public static final void SSS (double[] a, double[] b, double[] c, int n) {
    if (b.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(double)MMath.ModuloEuclid((double)a[i], (double)b[0]);
      }
    } else if (a.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(double)MMath.ModuloEuclid((double)a[0], (double)b[i]);
      }
    } else {
      for (int i=0; i<n; i++) {
        c[i]=(double)MMath.ModuloEuclid((double)a[i], (double)b[i]);
      }
    }
  }

  public final int process (float[][] v, int n, int mode) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (float[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(float[][],int,int)
   */
  public static final void SSS (float[] a, float[] b, float[] c, int n) {
    if (b.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(float)MMath.ModuloEuclid((double)a[i], (double)b[0]);
      }
    } else if (a.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(float)MMath.ModuloEuclid((double)a[0], (double)b[i]);
      }
    } else {
      for (int i=0; i<n; i++) {
        c[i]=(float)MMath.ModuloEuclid((double)a[i], (double)b[i]);
      }
    }
  }

  public final int process (int[][] v, int n, int mode) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  public final int process (int[][] v, int n, int mode, double scale) {
    switch (mode) {
      case SSS: SSS(v[0],v[1],v[2],n); return n;
      default:  return NOSUPPORT;
    }
  }

  /** Apply this operator to mode {@link Operator#SSS} data.
      @param a     The first  operand.
      @param b     The second  operand.
      @param c     The third  operand.
      @param n     The number of elements to operate on.
      @see #process(int[][],int,int)
   */
  public static final void SSS (int[] a, int[] b, int[] c, int n) {
    if (b.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(int)MMath.ModuloEuclid((double)a[i], (double)b[0]);
      }
    } else if (a.length == 1) {
      for (int i=0; i<n; i++) {
        c[i]=(int)MMath.ModuloEuclid((double)a[0], (double)b[i]);
      }
    } else {
      for (int i=0; i<n; i++) {
        c[i]=(int)MMath.ModuloEuclid((double)a[i], (double)b[i]);
      }
    }
  }

  public int getInputs() { return 2; }
  public int getOutputs() { return 1; }
}
