package nxm.sys.prim;

import nxm.sys.inc.InternalUseOnly;
import nxm.sys.lib.*;
import nxm.sys.net.dgnetsend;

/**
  sinkdg - uses dgnetsend class to write (Raw data) to Multicast or unicast datagrams

  @author MSZ / Neon Ngo
  @version $Id: sinkdg.java,v 1.21 2015/07/02 19:04:21 ntn Exp $
*/
public class sinkdg extends Primitive {
  private TextFile logfile;
  private dgnetsend dgsend = null;

  private static final int defaultport=18800,defaulttimeout=5000,defaultttl=1,defaultrtp=0;
  private static final int PLEN=2048;
  private byte[] inbuf=null;
  private DataFile hin=null;
  private short ptype=97;
  private int ssrc=1,blen,csrc0=0;
  private double freq=0.0,bw=0.0;

  private String host=null,hinformat=null;
  private int tl,port=defaultport,timeOut=defaulttimeout,ttl=defaultttl,rtp=defaultrtp;
  private boolean verbose=false,packetDebug=false,keepopen=true,opened=false,popened=false,rtptime=false,icein=false;
  /** output network packets from file/pipe data read? */
  private boolean outputPackets=true;

  public void setVerbose(boolean ptemp) { verbose=ptemp; }
  public void setRTPType(int rtptype) { ptype=(short)rtptype; }
  public void setRTPSSRC(int pssrc) { ssrc=pssrc; }
  public void setRTPCSRC(int pcsrc) { csrc0=pcsrc; }
  public void setSocketDebug(boolean ptemp) { if (dgsend!=null) try { dgsend.setVerbose(ptemp); } catch (Exception e) { e.printStackTrace(); } }
  public void setPacketDebug(boolean ptemp) { packetDebug=ptemp; if (dgsend!=null) try { dgsend.setPacketDebug(ptemp); } catch (Exception e) { e.printStackTrace(); } }
  public void setPTypeSSRC(int pptype,int pssrc) { ptype=(short)pptype; ssrc=pssrc; }
  public void setPTypeSRCs(int pptype,int pssrc,int pcsrc) { ptype=(short)pptype; ssrc=pssrc; csrc0=pcsrc; }
  public void setTimeOut(int ptimeout) { }
  public void setTTL(int pttl) { ttl=pttl; if (dgsend!=null) try { dgsend.setTTL(ttl); } catch (Exception e) { e.printStackTrace(); } }
  public String toString() { return "DG: Host="+host+" Port="+port; }
  public void setKeepOpen(boolean pkeep) { keepopen=pkeep; if (dgsend!=null) try { dgsend.setKeepOpen(pkeep); } catch (Exception e) { e.printStackTrace(); } }
  public void setHost(String phost) { host=phost; if (dgsend!=null) try { dgsend.setHost(phost); } catch (Exception e) { e.printStackTrace(); } }
  public String getHost() { return host; } // since 2.9.2
  public void setPort(int pport) { port=pport; if (dgsend!=null) try { dgsend.setPort(pport); } catch (Exception e) { e.printStackTrace(); } }
  public int getPort() { return port; }    // since 2.9.2
  public void setHostPort(String phost,int pport) {
    host=phost;
    port=pport;
    if (dgsend!=null) {
      try {
         dgsend.setHost(host);
         dgsend.setPort(port);
         if (!opened) open(); //does a close and then a reopen to create a new socket
      } catch (Exception e) { M.printStackTrace(e); }
    } else M.info("sinkdg: class not created ");
  }

  public dgnetsend getDgnetsend() { return dgsend; } // since 3.0.0

  @InternalUseOnly("since initial version in NeXtMidas 3.5.4 for debugging purposes only")
  public dgnetsend getDgsend() { return dgsend; }

  /** Turn on/off output of network packets for data read from file/pipe.
      @param val set to false to pause sending network packets, true to on/resume
      @since NeXtMidas 2.9.0
   */
  public void setOutputPackets (boolean val) {
    outputPackets = val;
  }
  /** Get current output network packets state.
      @return true for on/send, false for off/paused.
      @since NeXtMidas 2.9.0
   */
  public boolean getOutputPackets () { return outputPackets; }

  public int open () {
    try {
      if (dgsend==null) {
        dgsend = new dgnetsend();
        if (dgsend==null) { M.error("sinkdg: Could not open socket"); return FINISH; }
        if (MA.find("/VERBOSE")) verbose=true;
        if (MA.find("/PACKETS")) packetDebug=true;
        timeOut = MA.getL("/TIMEOUT",defaulttimeout);
        if (MA.find("/KEEPOPEN")) keepopen=true;
        ttl = MA.getL("/TTL",defaultttl);
        rtp = MA.getL("/RTP",defaultrtp);
        rtptime = MA.getState("/RTPTIME",rtptime);
        icein = MA.getState("/ICEIN",icein);
        ptype = MA.getI("/PTYPE",ptype);
        ssrc = MA.getL("/SSRC",ssrc);
        csrc0 = MA.getL("/CSRC",csrc0);
        tl = MA.getL("/TL",PLEN);
        String hostport=MA.getS("/HOSTPORT","");

        if (MA.find("/LOG")) {
          logfile = MA.getTextFile("/LOG");
          if (!logfile.open(TextFile.OUTPUT|TextFile.NOABORT)) {
            logfile.writeln("sinkdg log started at "+Time.tag());
          } else {
            M.warning("Could not find URL="+logfile.getURL());
            logfile = null;
          }
        }
        setVerbose(verbose);
        setPacketDebug(packetDebug);
        setTTL(ttl);
        if (verbose) {
          M.info("sinkdg: Verbose="+verbose+". PacketDebug="+packetDebug+". Keepopen="+keepopen);
          M.info("sinkdg: TTL="+ttl);
        }
        if (keepopen && host==null) {
          int i=hostport.indexOf(':');
          if (i>0) {
            host=hostport.substring(0,i).toLowerCase();
            port=Integer.parseInt(hostport.substring(i+1));
            M.info("sinkdg: HOSTPORT="+hostport);
          }
        }
      } // end if (dgsend==null)

      // 2010-12-07 NTN: moved the below 4 line outside the above if (dgsend==null) block
      //                 so that SINKDG can support pipe restarts - BUG 2300
      hin = MA.getDataFile ("IN","1000,2000","",0);
      popened=false;
      if (hin==null) M.info("Invalid Input File - 1000,2000 only");
      hinformat=hin.getFormat();

      if (verbose && host!=null) M.info("sinkdg: HOST,PORT="+host+","+port);
      if (host!=null && hin!=null && !popened) {
        String fname = MA.getS("IN");
        if (fname.length()>0) {
          M.info("Opening hin "+fname+","+hin.isOpen()+" for "+host);
          hin.open();
          hinformat=hin.getFormat();
          popened=true;
          inbuf = new byte[tl];
          keepopen=true;
        } else { M.info("sinkdg Invalid Input File"); hin.close(); hin=null; }
      }

      if (keepopen) {
        if (host!=null) { //this hasn't been set yet
          dgsend.open(host,port);
          opened=true;
          if (verbose) M.info("sinkdg: Host="+host+". Port="+port);
        } else M.info("sinkdg: host is null");
      } else {
        if (host!=null) { //this hasn't been set yet
          dgsend.open();
          opened=true;
          if (verbose) M.info("sinkdg: Host="+host+". Port="+port);
        }
      }
    } catch (Exception e) {
      throw new MidasException("sinkdg: open failed", e);
    }
    return NORMAL;
  }

  public int close () {
    try {
      if (logfile!=null) {
        logfile.writeln("sinkdg log stopped at "+Time.tag());
        logfile.close();
      }
      if (hin!=null) hin.close();
      if (dgsend!=null) dgsend.close();
    } catch (Exception e) {
      throw new MidasException("sinkdg: close failed", e);
    }
    return NORMAL;
  }

  /** @return 1 if sent successfully, 0 if unable to send, -1 for invalid input or error state */
  public int send (byte[] buf) {
    try {
      if (verbose) M.info("senddg: send(byte[] buf) Packet data="+new String(buf));
      if (buf==null || dgsend==null) return -1;
      boolean sent = dgsend.write(buf);
      if (!sent) return 0; // sent failed (since 3.3.0)
    } catch (Exception e) {
      throw new MidasException("sinkdg: send failed", e);
    }
    return 1;
  }

  /** @return 1 if sent successfully, 0 if unable to send, -1 for invalid input or error state */
  public int send (String data) {
    try {
      if (verbose) M.info("senddg: send(String data) Packet data="+data);
      if (data==null || dgsend==null) return -1;
      byte[] buf = data.getBytes();
      boolean sent = dgsend.write(buf);
      if (!sent) return 0; // sent failed (since 3.3.0)
    } catch (Exception e) {
      throw new MidasException("sinkdg: send failed", e);
    }
    return 1;
  }

  /** @return 1 if sent successfully, 0 if unable to send, -1 for invalid input or error state */
  public int send (String host, String port, String data) {
    try {
      if (verbose) M.info("senddg: send(host,port,data)->Packet data="+data);
      if (host==null || data==null || dgsend==null) return -1;
      int iPort = Integer.parseInt(port);
      if (iPort<0 || iPort>65535) return -1;
      boolean sent = dgsend.write(host,iPort,data);
      if (!sent) return 0; // sent failed (since 3.3.0)
    } catch (Exception e) {
      throw new MidasException("sinkdg: send failed", e);
    }
    return 1;
  }

  /** @return 1 if sent successfully, 0 if unable to send, -1 for invalid input or error state */
  public int sendRTP (byte[] buf, int bsize) {
    try {
      if (verbose) M.info("senddg: send(byte[] buf) Packet data="+new String(buf));
      if (buf==null || dgsend==null) return -1; // sent failed (invalid data or null dgnetsend library)
      boolean sent = dgsend.writeRTP(buf,bsize,ptype,ssrc,0,csrc0);
      if (!sent) return 0; // sent failed (since 3.3.0)
    } catch (Exception e) {
      throw new MidasException("sinkdg: sendRTP failed", e);
    }
    return 1; // sent successful
  }

  /** @return 1 if sent successfully, 0 if unable to send, -1 for invalid input or error state */
  public int sendRTP (String data) {
    try {
      if (verbose) M.info("senddg: send(String data) Packet data="+data);
      if (data==null || dgsend==null) return -1; // sent failed (invalid data or null dgnetsend library)
      byte[] buf = data.getBytes();
      boolean sent = dgsend.writeRTP(buf,buf.length,ptype,ssrc,0,csrc0);
      if (!sent) return 0; // sent failed (since 3.3.0)
    } catch (Exception e) {
      throw new MidasException("sinkdg: sendRTP failed", e);
    }
    return 1; // sent successful
  }

  public int process () {
    if (inbuf==null) return NOOP;

    blen=hin.read(inbuf,0,tl);
    if (blen==0) return NOOP;
    if (blen<0) return FINISH;
    if (outputPackets == false) return NORMAL;

    if (rtp>0) {
      if (rtptime) {
        if (icein) {
          dgsend.writeRTPICE(inbuf,blen,ptype,ssrc,csrc0); //takes the ice header from start of pipe
        } else {
          dgsend.writeRTPICE(inbuf,blen,hinformat,Time.current(),freq,bw,ptype,ssrc,0,csrc0);
        }
      } else {
        if (icein) dgsend.writeRTP(inbuf,blen-ICEPacket.headerLength,ptype,ssrc,ICEPacket.headerLength,csrc0); //skip iceheader
        else       dgsend.writeRTP(inbuf,blen,ptype,ssrc,0,csrc0);
      }
    } else dgsend.write(inbuf);
    return NORMAL;
  }

  public synchronized int processMessage (Message msg) {
    try {
      if (verbose) M.info("sinkdg: Processing a message: "+msg.name);
      if (msg.name.equals("OPENNET")) {
        if (keepopen) {
          String hostport=(String)msg.data;
          int i=hostport.indexOf(':');
          if (i>0) {
            host=hostport.substring(0,i).toLowerCase();
            port=Integer.parseInt(hostport.substring(i+1));
            dgsend.open(host,port);
            if (verbose) M.info("sinkdg: Host="+host+". Port="+port);
          } else {
            M.error("sinkdg: open failed: host:port wrong"+host+":"+port+". hostport="+hostport);
          }
        } else dgsend.open();
      } else if (msg.name.equals("SENDNET")) {
//      if (keepopen)  { dgsend.write(msg.data); }
//      else {
//        String hostport=msg.info;
//        int i=hostport.indexOf(':');
//        if (i>0) {
//          host=hostport.substring(0,i).toLowerCase();
//          port=Integer.parseInt(hostport.substring(i+1));
//          if (verbose) M.info("sinkdg: Host="+host+". Port="+port);
//          dgsend.write(host,port,msg.data);
//        }
//      }
      } else if (msg.name.equals("CLOSENET")) {
        dgsend.close();
      } else if (msg.name.equals("EXITNET")) {
        dgsend.close();
        return FINISH;
      }
    } catch (Exception e) {
      throw new MidasException("sinkdg: processMessage failed", e);
    }
    return NORMAL;
  }

  public int getSSRC() { return ssrc; }

}
