package nxm.sys.prim;

import nxm.sys.inc.*;
import nxm.sys.lib.*;

import java.util.List;

/**
  Takes input data from a file and outputs to a file/pipe.

  @author Beth Hassler
  @version $Id: sourcefile.java,v 1.10 2013/01/04 15:14:07 ntn Exp $
*/
public class sourcefile extends Primitive {

  public  double offout, off, stop[];
  public  int pause=0, replay;
  public  double startTime;
  public  DataFile hin, hout;

  private Data dbi,dbo;
  private double dur, slip;
  private double[] start;
  private int red,tl;
  private float[] fbuf;
  private byte[] inbuf;
  private boolean selfpace=true, reset=false;
  private double time, tdelay, tlast=0;
  private int idx, pass=0;

  private static final String replayList="ONESHOT,WAITTOP,ONETOP,CONTINUOUS";
  private static final int R_ONESHOT=1,R_WAITTOP=2,R_ONETOP=3,R_CONT=4;

  public int open() {
    // Open input file
    hin = MA.getDataFile("IN");
    hin.open();
    selfpace = hin.isFile();

    // Open output file
    hout = MA.getDataFile("OUT",hin,0);
    hout.open();

    // Get parameters
    idx = 0;
    start = stop = new double[1];
    start[idx] = off = MA.getD("START", 0.0);
    stop[idx]  = MA.getD("STOP", hin.size);
    dur = MA.getD("DURATION", 0.0);
    red = MA.getL("REDUCTION", 1);
    slip = MA.getD("SLIP", 0.0);

    // Setup processing
    tl = MA.getL("/TL",4096);
    tdelay = tl * hin.getXDelta();
    inbuf = new byte[tl];
    dbi = hin.getDataBuffer(tl);
    dbo = hout.getDataBuffer(tl);
    startTime = hin.getXStart();

    replay = MA.getChoice("/REPLAY",replayList,1);
    if (replay==R_WAITTOP || (replay==R_ONETOP && pass==1)) setPause(1);

    return NORMAL;
  }

  public void openFile(String name) {
    if(verbose)System.out.println("Sourcefile: Opening file "+name);
    hin = new DataFile();
    hin.init(M,name);
    hin.open();
    selfpace = hin.isFile();
    idx = 0;
    start = stop = new double[1];
    start[idx] = off = 0.0;
    stop[idx]  = hin.size;
    tdelay = tl * hin.getXDelta();
    dbi = hin.getDataBuffer(tl);
    dbo = hout.getDataBuffer(tl);
    startTime = hin.getXStart();
  }

  public int process() {

    if (pause==1) return NOOP;

    int n = hin.read(dbi,off,tl);
    if (n==0) {
      pass += 1;
      if (replay==R_ONESHOT) return FINISH;
      else return RESTART;
    }
    if (n<0) {
      pass += 1;
      if (replay==R_ONESHOT) return FINISH;
      else return RESTART;
    }
    off += (double)n;
    if (reset || off>stop[idx]) {
      if (idx<start.length-1) idx+=1; else idx=0;
      if (reset) idx=0;
      off=start[idx];
      reset = false;
    }

    if (selfpace) {
      time=Time.current();
      while ( (time-tlast) < (tdelay-.001) ) {
        Time.sleep(0.001);
        time=Time.current();
      }
      tlast = time;
    }

    dbo = dbi;
    int j = hout.write(dbo, n);
    offout = off*hin.getXDelta()+startTime;

    return NOOP;
  }

  /** Process a message */
  public int processMessage (Message msg) {
    if (msg.name.equals("OPENFILE")) {
      openFile((String)msg.data);
    } else if (msg.name.equals("CLOSEFILE")) {
      closeFile((String)msg.data);
    }
    return NORMAL;
  }

  public int close() {
    hin.close();
    hout.update();
    hout.close();
    return NORMAL;
  }

  public void setPause (int val) {
    pause = val;
  }

  public void setStart (double val) {
    start = new double[1];
    idx = 0;
    off = (val-startTime)/hin.getXDelta();
    offout = off*hin.getXDelta()+startTime;
    start[idx] = off;
  }

  public void setStop (double val) {
    stop = new double[1];
    idx = 0;
    if (val == -1.0) {
      stop[idx] =  hin.size;
    } else {
      stop[idx] = (val-startTime)/hin.getXDelta();
    }
  }

  public void setStartStops (Table t) {
    int size = t.size();
    if (size>0) {
      start = new double[size];
      stop = new double[size];
      for (Table.Iterator ti=t.iterator(); ti.getNext();) {
        Table tmpt = t.getTable(ti.key);
        int indx = tmpt.getI("INDX");
        if (indx<size) {
          start[indx] = (tmpt.getD("START")-startTime)/hin.getXDelta();
          stop[indx]  = (tmpt.getD("STOP")-startTime)/hin.getXDelta();
        } else
          M.warning("Error in Sourcefile - index out of bounds");
      }
      reset = true;
    }
  }

  public void setStarts (List<?> v) {
    if (verbose) System.out.println("Sourcefile: TBD - STARTS: "+v.toString());
  }

  public void setStops (List<?> v) {
    if (verbose) System.out.println("Sourcefile: TBD - STOPS: "+v.toString());
  }

  public void closeFile(String name) {
    hin.close();
  }

  public int restart() {
    super.restart();
    return NORMAL;
  }

  public int getReplay() { return replay; }

  public void setReplay (int replay) {
    this.replay = replay;
  }

}
