package nxm.ucl.libg;

import java.util.Calendar;

/** A class that creates a simple text/combo box that allows the user to
    type in and/or spin a year value.  The selected year is used to update
    the given calendar, which will add the specific year value to its
    currently selected date, and re-render the calendar.

    @author crs
    @version $Id: MJYearPicker.java,v 1.2 2012/12/28 20:03:17 ntn Exp $
    @since NeXtMidas 2.9.1
 */
public class MJYearPicker extends MJSpinField {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012122810331L; // for Serializable interface

  protected int oldYear;
  protected int startYear;
  protected int endYear;

  /** GDatePicker objects **/
  private MJMonthPicker monthPicker;
  private MJDayPicker   dayPicker;

  /**
    Default constructor.
   */
  public MJYearPicker () {
    setName("MJYearPicker");
    Calendar calendar = Calendar.getInstance();
    setMinimum(calendar.getMinimum(Calendar.YEAR));
    setMaximum(calendar.getMaximum(Calendar.YEAR));
    setValue(calendar.get(Calendar.YEAR));
  }

  /**
    Sets the year on this component and fires an event, "year", to any
    listening component that the year has changed.
    @param year The new year value
   */
  public void setYear (int year) {
    super.setValue(year, true, false);

    if (dayPicker != null) {
      dayPicker.setYear(year);
    }
    spinner.setValue(new Integer(value));
    firePropertyChange("year", oldYear, value);
    oldYear = value;
  }

  /**
    Returns the current year value.
    @return Integer year value
   */
  public int getYear () {
    return super.getValue();
  }

  /**
    Sets the year value.
    @param value The new year value
   */
  public void setValue (int value) {
    setYear(value);
  }

  /**
    Sets the start year, or minimum year.
    @param startYear The minimum start year
   */
  public void setStartYear (int startYear) {
    setMinimum(startYear);
  }

  /** Accessory Methods **/

  /**
    Sets this object's MonthPicker reference.
    @param monthPicker The MonthPicker object to reference.
   */
  public void setMonthPicker (MJMonthPicker monthPicker) {
    this.monthPicker = monthPicker;
  }

  /**
    Sets this object's DayPicker reference.
    @param dayPicker The DayPicker object to reference.
   */
  public void setDayPicker (MJDayPicker dayPicker) {
    this.dayPicker = dayPicker;
  }

}
