package nxm.ucl.libg;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JFrame;
import javax.swing.JPanel;

import nxm.sys.inc.InternalUseOnly;
import nxm.sys.lib.MessageQueue;
import nxm.sys.lib.Midas;
import nxm.sys.lib.Results;
import nxm.sys.libg.MidasBean;
import nxm.sys.libg.MidasDisplay;

/** Used to duplicate BUG 1905
    @author Wayne Dengel
    @since NeXtMidas 2.7.2
    @version $Id: WHTabPanel.java,v 1.3 2014/12/15 15:57:44 ntn Exp $
 */
@InternalUseOnly("Since NeXtMidas 3.5.2")
public class WHTabPanel extends JPanel {
  /** NOTE: this should be changed if the structure/fields of this class changes. */
  private static final long serialVersionUID = 2012122803331L; // for Serializable interface

  protected Color         bg            = null;
  protected Color         fg            = null;
  public    JFrame        jFrame        = null;
  protected Results       MR            = null;
  protected MessageQueue  MQ            = null;
  protected MidasDisplay  rasterDisplay = new MidasDisplay();
  protected MidasBean     rasterBean    = null;
  protected Midas         M             = null;
  protected String        name          = null;
  protected int           tabID         = 0;
  public    boolean       docked        = true;

  public WHTabPanel (Midas M, MessageQueue mq, Results mr, Color bg, Color fg, String name, int tabID) {
    this.M     = M;
    this.bg    = bg;
    this.fg    = fg;
    this.MR    = mr;
    this.MQ    = mq;
    this.name  = name;
    this.tabID = tabID;
  }

  public void redock () {
    jFrame.setVisible(false);
    docked = true;
  }

  public GraphicsConfiguration undock (int screen, int x) {
    if (screen > -1) {
      GraphicsConfiguration gc = getDisplay(screen);
      if (jFrame != null) {
        jFrame.setVisible(false);
        jFrame.dispose();
        jFrame = null;
      }
      jFrame = new JFrame(this.name,gc);
      jFrame.add(this, BorderLayout.CENTER);
      jFrame.setSize(x, 855);
      Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
      Dimension frameSize  = jFrame.getSize();
      if (frameSize.height > screenSize.height) frameSize.height = screenSize.height;
      if (frameSize.width  > screenSize.width)  frameSize.width = screenSize.width;
      jFrame.setLocation( (screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);
      jFrame.addWindowListener(new WindowAdapter() { public void windowClosing(WindowEvent e) {redock();}});
      jFrame.setVisible(true);
      docked = false;
      return gc;
    }
    return null;
  }

  private GraphicsConfiguration getDisplay (int display) {
    GraphicsConfiguration gconf = null;
    if (display > -1) {
      try {
        gconf = GraphicsEnvironment.getLocalGraphicsEnvironment().getScreenDevices()[display].getConfigurations()[0];
      } catch (Exception e) {
        M.warning("WHTabPaneL: Unable to connect to display '" +display+"': "+ e);
      }
    }
    return gconf;
  }

  public int getTabID () { return tabID; }

  public String getName() { return name; }

}
