// -*- C++ -*-
#ifndef XM_FINTRINSICS_H_

// C++ versions of common Fortran intrinsics, to ease primitive conversion.
//
// Author:  Based on X-Midas 4.6.1 fintrinsics.h
// Version: $Id: fintrinsics.h,v 1.1 2009/08/11 00:25:53 ntn Exp $

#include <cmath>    // std::floor, std::ceil

// Midas includes
#include "cdefs.h"  // int_*, real_*

// ///// Functions that duplicate needed Fortran elemental functions
// ///// for integer truncation and rounding

// Truncate the fractional part of the real value passed in, returning
// an integer.

inline int_4 INT(real_4 x) { return int_4(x); }
inline int_4 INT(real_8 x) { return int_4(x); }
inline int_4 JINT(real_4 x) { return int_4(x); }

inline int_2 IINT(real_4 x) { return int_2(x); }
inline int_2 IIFIX(real_4 x) { return int_2(x); }
inline int_2 HFIX(real_4 x) { return int_2(x); }

inline int_4 IFIX(real_4 x) { return int_4(x); }
inline int_4 JFIX(real_4 x) { return int_4(x); }
inline int_4 JFIX(real_8 x) { return int_4(x); }

inline int_8 KIFIX(real_4 x) { return int_8(x); }
inline int_8 KINT(real_4 x) { return int_8(x); }

inline int_2 IIDINT(real_8 x) { return int_2(x); }
inline int_4 IDINT(real_8 x) { return int_4(x); }
inline int_4 JIDINT(real_8 x) { return int_4(x); }
inline int_8 KIDINT(real_8 x) { return int_8(x); }


// Truncate the fractional part of the real value passed in, returning
// a real number.

inline real_4 AINT(real_4 x) { return (x > 0) ? std::floor(x) : std::ceil(x); }
inline real_8 DINT(real_8 x) { return (x > 0) ? std::floor(x) : std::ceil(x); }


// Round the real value passed in to the nearest integer, returning an
// integer.  The nearest integer to 12.5 is 13, the nearest integer to
// -12.5 is -13.

inline int_2 ININT(real_4 x) { return  x > 0 ?
				 int_2(x + 0.5) : int_2(x - 0.5); }
inline int_4 NINT(real_4 x) { return  x > 0 ?
				int_4(x + 0.5) : int_4(x - 0.5); }
inline int_4 JNINT(real_4 x) { return  x > 0 ?
				 int_4(x + 0.5) : int_4(x - 0.5); }
inline int_8 KNINT(real_4 x) { return  x > 0 ?
				 int_8(x + 0.5) : int_8(x - 0.5); }

inline int_2 IIDNNT(real_8 x) { return  x > 0 ?
				  int_2(x + 0.5) : int_2(x - 0.5); }
inline int_4 IDNINT(real_8 x) { return  x > 0 ?
				  int_4(x + 0.5) : int_4(x - 0.5); }
inline int_4 JIDNNT(real_8 x) { return  x > 0 ?
				  int_4(x + 0.5) : int_4(x - 0.5); }
inline int_8 KIDNNT(real_8 x) { return  x > 0 ?
				  int_8(x + 0.5) : int_8(x - 0.5); }


// Round the real value passed in to the nearest integer, returning a
// real number.  The nearest integer to 12.5 is 13, the nearest
// integer to -12.5 is -13.

inline real_4 ANINT(real_4 x)
{
  return (x > 0) ? std::floor(x + 0.5f) : std::ceil(x - 0.5f);
}

inline real_8 DNINT(real_8 x)
{
  return (x > 0) ? std::floor(x + 0.5) : std::ceil(x - 0.5);
}

inline bool BTEST(int_4 x, int_4 y)
{
  int_4 tpow = (1 << y);
  return (x & tpow) == tpow;
}


inline int_4 IBSET(int_4 bitv, int_4 which_bit)
{
  return bitv | (1 << which_bit);
}


#define XM_FINTRINSICS_H_
#endif // XM_FINTRINSICS_H_
